import javax.swing.border.*;
import java.awt.*;
import java.awt.event.*;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.swing.*;
import javax.swing.Timer;
import java.io.*;
import java.util.*;

/**
 * <hr>
 * 
 * <table width="100%" summary="ScanningKeyboardExperiment title">
 * <tr style="background-color:#cccccc">
 * <td align="center">
 * <h1>ScanningKeyboardExperiment</h1>
 * </table>
 * <p>
 * 
 * <h2>Summary</h2>
 * 
 * <ul>
 * <li>Experiment software for text entry using a scanning keyboard
 * <P>
 * <li>Only the Spacebar is used for input
 * <P>
 * <li>Several keyboard layouts are supported, some including word prediction and completion
 * <P>
 * <li>A custom keyboard can be defined and read from a text file
 * </ul>
 * <p>
 * 
 * <h2>Related References</h2>
 * 
 * <ul>
 * <li><a href="http://www.yorku.ca/mack/icchp2012.html">"Modeling Text Input for Single-switch
 * Scanning"</a>, by MacKenzie (<i>ICCHP 2012</i>). Describes modeling, as the title suggests, and
 * also presents and examines the metric "scansteps per character" (SPC).
 * </ul>
 * 
 * <h2>Running the Experiment Software</h2>
 * 
 * <a href="http://www.yorku.ca/mack/HCIbook/Running/">Click here</a> for instructions on
 * launching/running the application.
 * <p>
 * 
 * <h2>Setup Parameters</h2>
 * 
 * Upon launching, the program presents a setup dialog:
 * <p>
 * 
 * <center> <img src="ScanningKeyboardExperiment-1.jpg" alt="ScanningKeyboardExperiment-1">
 * </center>
 * <p>
 * 
 * The default parameter settings are read from a configuration file called
 * <a href="ScanningKeyboardExperiment.cfg"> <code>ScanningKeyboardExperiment.cfg</code></a>. This
 * file is created automatically when the application is launched for the first time. The default
 * parameter settings may be changed through the setup dialog. The setup parameters are as follows:
 * <p>
 * 
 * <blockquote>
 * <table border="1" cellspacing="0" cellpadding="6" summary="setup paramters">
 * <tr style="background-color:#cccccc">
 * <th align="center">Parameter
 * <th align="center">Description
 * 
 * <tr valign="top">
 * <td>Participant code
 * <td>Identifies the current participant. This is used in forming the names for the output data
 * files. Also, the sd2 output data file includes a column with the participant code.
 * <p>
 * 
 * <tr valign="top">
 * <td>Condition code
 * <td>An arbitrary code used to associate a test condition with this invocation. This parameter
 * might be useful if the software is used in an experiment where a condition is not inherently part
 * of the application (e.g., Gender &rarr; male, female, Input Device &rarr; gamepad, keyboard). The
 * condition code is used in forming the name for the output data file. Also, the sd2 output data
 * file contains a column with the condition code.
 * <p>
 * 
 * <tr valign="top">
 * <td>Session code
 * <td>Identifies the session. Useful if testing proceeds over multiple sessions to gauge the
 * progression of learning.
 * <p>
 * 
 * Note: The setup dialog does not include an entry for "Block code". The block code is generated
 * automatically by the software.
 * <p>
 * 
 * <tr valign="top">
 * <td>Number of phrases
 * <td>Specifies the number of phrases presented to the participant in the current block.
 * <p>
 * 
 * <tr valign="top">
 * <td>Scanning interval
 * <td>The time in milliseconds for each scan step. Applies both for row scanning and column
 * scanning.
 * <p>
 * 
 * <tr valign="top">
 * <td>Scan delay
 * <td>A delay time in milliseconds that is inserted each time the Spacebar key is pressed. Applies
 * both when Spacebar is pressed during row scanning (to switch to column scanning) and when
 * Spacebar is pressed during column scanning (to select a character or word and then revert to row
 * scanning). The added delay gives the user a little extra time to adjust to the new scanning mode.
 * <p>
 * 
 * <tr valign="top">
 * <td>Phrases file
 * <td>Specifies a file containing phrases of text to be presented to participants for entry.
 * Phrases are drawn from the file at random. Typically, <a href="phrases2.txt">phrases2.txt</a> is
 * used. This is the phrase set published by MacKenzie and Soukoreff in 2003
 * (<a href="http://www.yorku.ca/mack/chi03b.html">click here</a>).
 * <p>
 * 
 * <tr valign="top">
 * <td>Show presented text during input
 * <td>A checkbox item that determines whether or not the presented text is visible during entry of
 * a phrase. Either way, the text phrase appears at the beginning of a trial. If this option is
 * unchecked, the phrase will disappear when the first keystroke is entered.
 * <p>
 * 
 * 
 * <tr valign="top">
 * <td>Audio on
 * <td>If on, audible "clicks" are heard during row and column scan steps and when the Spacebar is
 * pressed.
 * <p>
 * 
 * <tr valign="top">
 * <td>Keyboard
 * <td>A drop-down list to select the scanning keyboard for the current invocation. The options
 * currently supported are as follows: (click to enlarge)
 * <p>
 * 
 * <center> <b>Qwerty</b><br>
 * <a href="ScanningKeyboardExperiment-qwerty.jpg"><img src="ScanningKeyboardExperiment-qwerty.jpg"
 * width=300 alt="ScanningKeyboardExperiment-qwerty"></a>
 * <p>
 * 
 * <b>Qwerty_WP</b><br>
 * <a href="ScanningKeyboardExperiment-qwertywp.jpg"><img src=
 * "ScanningKeyboardExperiment-qwertywp.jpg" width=300 alt=
 * "ScanningKeyboardExperiment-qwertywp"></a>
 * <p>
 * 
 * <b>ABC1</b><br>
 * <a href="ScanningKeyboardExperiment-abc1.jpg"><img src="ScanningKeyboardExperiment-abc1.jpg"
 * width=150 alt="ScanningKeyboardExperiment-abc1"></a>
 * <p>
 * 
 * <b>ABC2</b><br>
 * <a href="ScanningKeyboardExperiment-abc2.jpg"><img src="ScanningKeyboardExperiment-abc2.jpg"
 * width=150 alt="ScanningKeyboardExperiment-abc2"></a>
 * <p>
 * 
 * <b>ABC</b><br>
 * <a href="ScanningKeyboardExperiment-abc.jpg"><img src="ScanningKeyboardExperiment-abc.jpg"
 * width=300 alt="ScanningKeyboardExperiment-abc"></a>
 * <p>
 * 
 * <b>ABC_WP</b><br>
 * <a href="ScanningKeyboardExperiment-abcwp.jpg"><img src="ScanningKeyboardExperiment-abcwp.jpg"
 * width=300 alt="ScanningKeyboardExperiment-abc"></a>
 * <p>
 * 
 * <b>Optimized</b><br>
 * <a href="ScanningKeyboardExperiment-optimized.jpg"><img src=
 * "ScanningKeyboardExperiment-optimized.jpg" width=150 alt=
 * "ScanningKeyboardExperiment-optimized"></a>
 * <p>
 *  
 * <b>Optimized_WP</b><br>
 * <a href="ScanningKeyboardExperiment-optimizedwp.jpg"><img src=
 * "ScanningKeyboardExperiment-optimizedwp.jpg" width=150 alt=
 * "ScanningKeyboardExperiment-optimizedwp"></a>
 * <p>
 * 
 * <b>Custom (select below)</b> </center>
 * 
 * <b>Note:</b> The optimized layout above is one of many such designs aimded at reducing the
 * average number of scan steps required for each character of text produced. Consult the reference
 * at the top of this API for details on this and other optimized layouts for scanning keyboards.
 * <p>
 * 
 * <tr valign="top">
 * <td>Custom keyboard
 * <td>If the "Custom" option is chosen above, this button is clicked to launch a file chooser to
 * select a text file containing the definition of a custom keyboard.
 * <p>
 * 
 * Each key in a scanning keyboard is defined by seven parameters:
 * <p>
 * 
 * <ul>
 * <li>Symbol &mdash; the text that appears on the key (Exception: "WORD" is used if the key is used
 * to present candidate words for keyboards using word prediction and completion)
 * <li>rowIdx &mdash; the row index for the corresponding key
 * <li>colIdx &mdash; the column index for the corresponding key
 * <li>x &mdash; the relative x position of the key
 * <li>y &mdash; the relative y position of the key
 * <li>width &mdash; the relative width of the key
 * <li>height &mdash; the relative height of the key
 * </ul>
 * <p>
 * 
 * Note: The last four values above are multiplied by default key width and key height constants
 * that are embedded in the app.
 * <p>
 * 
 * <a href="custom-QWERTY_WP.txt">Click here</a> to see an example custom keyboard. As a learning
 * aide, this custom keyboard is the same as the Qwerty_WP keyboard that appears in the Keyboard
 * drop-down list.
 * <p>
 * 
 * <tr valign="top">
 * <td>Keyboard scale
 * <td>Selects the scale of the keyboard relative to the default key width and key height constants
 * embedded in the application.
 * <p>
 * 
 * </table>
 * </blockquote>
 * <p>
 * 
 * <h2>Program Operation</h2>
 * 
 * Upon clicking OK in the setup dialog, the application launches:
 * <p>
 * 
 * <center> <img src="ScanningKeyboardExperiment-2.jpg" alt="ScanningKeyboardExperiment-2">
 * </center>
 * <p>
 * 
 * A phrase is drawn at random from the specified phrases file and presented for input. After
 * pressing the Spacebar, scanning begins, row by row. The timing of user input begins when scanning
 * begins (after the initial press of the Spacebar key).
 * <p>
 * 
 * When the row containing the desired letter is highlighted (below, left), the user presses the
 * Spacebar to begin scanning within the row. This begins column scanning. When the key bearing the
 * desired letter is highlighted (below, center), the user presses the Spacebar again to select the
 * letter. The letter is added to the transcribed text phrase and scanning begins anew (below,
 * right). Note that it takes two presses of the Spacebar to enter each text character.
 * <p>
 * 
 * <center>
 * <a href="ScanningKeyboardExperiment-3.jpg"><img src="ScanningKeyboardExperiment-3.jpg" width=250
 * alt="ScanningKeyboardExperiment-3"></a>
 * <a href="ScanningKeyboardExperiment-4.jpg"><img src="ScanningKeyboardExperiment-4.jpg" width=250
 * alt="ScanningKeyboardExperiment-4"></a>
 * <a href="ScanningKeyboardExperiment-5.jpg"><img src="ScanningKeyboardExperiment-5.jpg" width=250
 * alt="ScanningKeyboardExperiment-5"></a> </center>
 * <p>
 * 
 * After entering the complete phrase, the user inputs "Enter" (by pressing the Spacebar when the
 * Enter key is highlighted). A pop dialog appears, showing the presented and transcribed text
 * phrases and some user performance measures. Below is an example.
 * <p>
 * 
 * <center> <img src="ScanningKeyboardExperiment-6.jpg" alt="ScanningKeyboardExperiment-6">
 * </center>
 * <p>
 * 
 * The program terminates after entry of the last phrase in a block.
 * <p>
 * 
 * Some of the keyboards (See the setup parameter for "Keyboard", above) include word prediction and
 * completion. In this case, a series of keys are used to display candidate words as entry proceeds.
 * The candidates are the most likely words matching the current word stem, based on an embedded
 * word-frequency list, or dictionary. The dictionary contains about 10,000 words. Note that at the
 * beginning of a phrase and between words the candidates are simply the most common words in
 * English (since the word stem is empty).
 * <p>
 * 
 * <h2>Output data files</h2>
 * 
 * At the end of each block, data are written to two "summary data" (sd) files. The "sd1" file
 * contains data on a per-character basis, one line per character. Each line contains three values:
 * (i) the character, command, or word entered, (ii) the timestamp in milliseconds, and (iii) the
 * number of scan steps. The "sd2" file contains data on a per-phrase basis, one line per phrase.
 * Example data files follow:
 * <p>
 * 
 * <ul>
 * <li><a href="ScanningKeyboardExperiment-sd1-example.txt">sd1 example</a>
 * <li><a href="ScanningKeyboardExperiment-sd2-example.txt">sd2 example</a>
 * </ul>
 * <p>
 * 
 * 
 * The data are full-precision, comma-delimited. Importing into a spreadsheet application provides a
 * convenient method to examine the data and do follow-on analyses. Below is an example of how the
 * data in the sd2 file might look after inporting into Microsoft Excel: (click to enlarge)
 * <p>
 * 
 * <center>
 * <a href="ScanningKeyboardExperiment-7.jpg"><img src="ScanningKeyboardExperiment-7.jpg" width=800
 * alt="ScanningKeyboardExperiment-7"></a> </center>
 * <p>
 * 
 * Several user performance measures in the sd2 file are potentially useful as dependent variables
 * in a user study. <i>Speed</i> (wpm) and <i>Error Rate</i> (%) are well known in research on text
 * entry methods. The metric called <i>SPC</i> (scan steps per character) is computed as
 * <p>
 * 
 * <center> <code>scan_steps / number_of_characters_in_transcribed_text</code> </center>
 * <p>
 * 
 * The numerator is the actual number of scan steps logged during entry of the current phrase. SPC
 * is the scanning-keyboard equivalent of KSPC (keystrokes per character), which is commonly used
 * for keyboards using physical keys for input.
 * <p>
 * 
 * <i>Efficiency</i> is computed as
 * <p>
 * 
 * <center> <code>scan_steps_min / scan_steps x 100</code> </center>
 * <p>
 *
 * Here, the numerator is the minimum number of scan steps possible for the current text phrase. If
 * Efficiency = 100%, the phrase was inputted with each character or word entered correctly and at
 * the earliest opportunity (i.e., with no errors that were corrected and with no missed
 * opportunities). If Efficiency &lt; 100%, the user either made one or more errors that were
 * corrected or missed some opportunities to enter the desired character or word at the earliest
 * opportunity. Note that the Efficiency measure is only meaningful if the transcribed text is error
 * free.
 * <p>
 * 
 * Additional measures include <i>Presses_of_Bksp</i>, <i>Presses_of_Word_Delete</i>,
 * <i>Row_wrap_arounds</i>, and <i>Col_wrap_arounds</i>. Presses_of_Bksp and Presses_of_Word_Delete
 * are the number of times the named key was pressed during input of the current phrase. Obviously,
 * Bksp is used to correct errors. Word_Delete (if present) is pressed to delete the last word
 * entered. This is particularly useful with keyboards using word prediction and word completion.
 * Having entered the wrong word, the error is more quickly corrected using the Word_Delete key
 * rather using multiple presses of the Bksp key.
 * <p>
 * 
 * The Row_wrap_arounds and Col_wrap_arounds measures indicate the number of times scanning wrapped
 * around, either during row scanning or column scanning, respectively. When scanning wraps around,
 * it means the user missed an opportunity, either to select a row or to select a column (character,
 * word, or command). Missed opportunities indicate inefficient input and may occur because scanning
 * is too fast for the user to make a selection, given the previous selection.
 * <p>
 * 
 * Actual output files use "ScanningKeyboardExperiment" as the base filename. This is followed by
 * the participant code, the condition code, the keyboard code, the session code, and the block
 * code, for example, <code>ScanningKeyboardExperiment-P01-C01-ABC_WP-S01-B01.sd1</code>.
 * <p>
 * 
 * <h2>Miscellaneous</h2>
 * 
 * When using this program in an experiment, it is a good idea to terminate all other applications
 * and disable the system's network connection. This will maintain the integrity of the data
 * collected and ensure that the program runs without hesitations.
 * <p>
 * 
 * @author Scott MacKenzie, 2016-2022
 */
public class ScanningKeyboardExperiment
{

	final static String APP_NAME = "ScanningKeyboardExperiment";
	final static int PARAMETERS = 12; // must equal the number of parameters defined below

	// setup parameters (1st value is default)
	final static String[] PARTICIPANT_CODES =
	{ "P00", "P01", "P02", "P03", "P04", "P05", "P06", "P07", "P08", "P09", "P10", "P11", "P12",
			"P13", "P14", "P15", "P16", "P17", "P18", "P19", "P20", "P21", "P22", "P23", "P24",
			"P25", "P26", "P27", "P28", "P29", "P30" };
	final static String[] CONDITION_CODE =
	{ "C00" };
	final static String[] SESSION_CODES =
	{ "S00", "S01", "S02", "S03", "S04", "S05", "S06", "S07", "S08", "S09", "S10", "S11", "S12",
			"S13", "S14", "S15", "S16", "S17", "S18", "S19", "S20", "S21", "S22", "S23", "S24",
			"S25", "S26", "S27", "S28", "S29", "S30" };
	final static String[] NUMBER_OF_PHRASES =
	{ "3", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10" };
	final static String[] SCAN_INTERVAL =
	{ "700", "100", "200", "300", "400", "500", "600", "700", "800", "900", "1000", "1100", "1200",
			"1300", "1400", "1500", "1600", "1700", "1800", "1900", "2000" };
	final static String[] SCAN_DELAY =
	{ "200", "100", "200", "300", "400", "500", "600", "700", "800", "900", "1000" };
	final static String[] PHRASES_FILE =
	{ "phrases2.txt" };
	final static String[] SHOW_PRESENTED =
	{ "true" };
	final static String[] AUDIO_ON =
	{ "true" };
	final static String[] KEYBOARD =
	{ "Qwerty", "Qwerty_WP", "ABC1", "ABC2", "ABC", "ABC_WP", "OPTIMIZED", "OPTIMIZED_WP",
			"Custom (select below)" };
	final static String[] KEYBOARD_DEF_FILE =
	{ null };
	final static String[] KEYBOARD_SCALE =
	{ "1.0" };

	// identify other files needed (used when executing from the .jar file for the 1st time)
	final static String[] OTHER_FILES =
	{ "phrases2.txt", "helloworld.txt", "quickbrownfox.txt", "tickRowScanning.wav",
			"tickColScanning.wav", "switch.wav", "d1phrases2-wordfreq.txt" };

	final static String[] KEYBOARD_CODE =
	{ "QWERTY", "QWERTY_WP", "ABC1", "ABC2", "ABC", "ABC_WP", "OPTIMIZED", "OPTIMIZED_WP",
			"CUSTOM" };

	public static void main(String[] args) throws IOException
	{
		// This change was copied from GoFitts.java

		// use look and feel for "this" system
		// try
		// {
		// UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		// } catch (Exception e)
		// {
		// }

		/*
		 * This was a time consuming bug to resolve. For some reason, the color buttons in the setup
		 * dialog stopped showing the color in the button background. I'm not sure exactly when or
		 * why this happened, but it was probably the result of a JRE upgrade, or something. It
		 * seems the problem was caused by a change in the "WindowsLookAndFeel", which was the
		 * default L&F returned above when calling the getSystemLookAndFeelClassName method. Below,
		 * I am explicitly using a different L&F, rather than using the system default. Everything
		 * is working fine now; i.e., the colors appear in the background in the JButtons for the
		 * color options. NOTE: I haven't tested this on other machines, such as MacOS machines.
		 * (S.M., Oct 19, 2020)
		 */

		try
		{
			UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");
			// UIManager.setLookAndFeel("javax.swing.plaf.metal.NimbusLookAndFeel");
			// UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");
			// UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
			// UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsClassicLookAndFeel");
		} catch (Exception e)
		{
		}

		/*
		 * Define the default parameters settings. These are only used if the .cfg file does not
		 * exist, such as when the application is launched for the 1st time.
		 */
		String[] defaultParameters = new String[PARAMETERS];
		defaultParameters[0] = PARTICIPANT_CODES[0];
		defaultParameters[1] = CONDITION_CODE[0];
		defaultParameters[2] = SESSION_CODES[0];
		defaultParameters[3] = NUMBER_OF_PHRASES[0];
		defaultParameters[4] = SCAN_INTERVAL[0];
		defaultParameters[5] = SCAN_DELAY[0];
		defaultParameters[6] = PHRASES_FILE[0];
		defaultParameters[7] = SHOW_PRESENTED[0];
		defaultParameters[8] = AUDIO_ON[0];
		defaultParameters[9] = KEYBOARD[0];
		defaultParameters[10] = KEYBOARD_DEF_FILE[0];
		defaultParameters[11] = KEYBOARD_SCALE[0];

		/*
		 * Instantiate a Configuration object. The configuration parameters are (a) read from the
		 * .cfg file or (b) the passed default parameters if the .cfg file does not exist.
		 */
		Configuration c = Configuration.readConfigurationData(APP_NAME, defaultParameters,
				OTHER_FILES);
		if (c == null)
		{
			System.out.println("Error reading configuration data from " + APP_NAME + ".cfg!");
			System.exit(0);
		}

		// use the app's parameters to define the SetupItemInfo objects needed to create the setup
		// dialog.
		SetupItemInfo[] sii = new SetupItemInfo[PARAMETERS];
		sii[0] = new SetupItemInfo(SetupItem.COMBO_BOX, "Participant code ", PARTICIPANT_CODES);
		sii[1] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Condition code ", CONDITION_CODE);
		sii[2] = new SetupItemInfo(SetupItem.COMBO_BOX, "Session code ", SESSION_CODES);
		sii[3] = new SetupItemInfo(SetupItem.COMBO_BOX, "Number of phrases ", NUMBER_OF_PHRASES);
		sii[4] = new SetupItemInfo(SetupItem.COMBO_BOX, "Scan interval ", SCAN_INTERVAL);
		sii[5] = new SetupItemInfo(SetupItem.COMBO_BOX, "Scan delay ", SCAN_DELAY);
		sii[6] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Phrases file ", PHRASES_FILE);
		sii[7] = new SetupItemInfo(SetupItem.CHECK_BOX, "Show presented text during input ",
				SHOW_PRESENTED);
		sii[8] = new SetupItemInfo(SetupItem.CHECK_BOX, "Audio on ", AUDIO_ON);
		sii[9] = new SetupItemInfo(SetupItem.COMBO_BOX, "Keyboard ", KEYBOARD);
		sii[10] = new SetupItemInfo(SetupItem.FILE_CHOOSER, "Custom keyboard ", KEYBOARD_DEF_FILE);
		sii[11] = new SetupItemInfo(SetupItem.TEXT_FIELD, "Keyboard scale ", KEYBOARD_SCALE);

		/*
		 * Create and show the setup dialog. The user can make changes to the default or existing
		 * configuration parameters
		 */
		Setup s = new Setup(null, c, APP_NAME, sii);
		s.showSetup(null);

		// launch application GUI
		ScanningKeyboardExperimentGui screen = new ScanningKeyboardExperimentGui(c);
		screen.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		screen.setTitle(APP_NAME);
		screen.pack();

		// put application in centre of screen
		int w = screen.getWidth();
		int h = screen.getHeight();
		Toolkit t = Toolkit.getDefaultToolkit();
		Dimension d = t.getScreenSize();
		screen.setLocation((d.width - w) / 2, (d.height - h) / 2);
		screen.setVisible(true);
	}
}

// =================================================
// ScanningKeyboardExperimentGui (main class for UI)
// =================================================

class ScanningKeyboardExperimentGui extends JFrame implements KeyListener, ScanningKeyboardListener
{
	// the following avoids a "warning" with Java 1.5.0 complier (?)
	static final long serialVersionUID = 42L;

	final String SD1_HEADER = "";
	final String SD2_HEADER = "App,Participant,Condition,Keyboard,Session,Block,Scan_interval,Scan_delay,"
			+ "Min_scan_steps,Scan_steps,Characters,Presses_of_Bksp,Presses_of_Word_Delete,Row_wrap_arounds,Col_wrap_arounds,Time,MSD,Speed_(wpm),Error_rate_(%),SPC,Efficiency_(%)\n";
	final String HIT_SPACEBAR_MESSAGE_TEXT = "Hit Spacebar to begin scanning";
	final Font F14 = new Font("sansserif", Font.PLAIN, 14);
	final KeyStroke SPACEBAR = KeyStroke.getKeyStroke(KeyEvent.VK_SPACE, 0);
	final String DICTIONARY_FILE = "d1phrases2-wordfreq.txt";

	// mode
	final int QWERTY = 100;
	final int QWERTY_WP = 110;
	final int ABC1 = 200;
	final int ABC2 = 300;
	final int ABC = 350;
	final int ABC_WP = 400;
	final int OPTIMIZED = 500;
	final int OPTIMIZED_WP = 550;
	final int CUSTOM = 600;

	// define a mode array (NOTE: values must be in same order as in setup dialog)
	final int[] MODE_ARRAY =
	{ QWERTY, QWERTY_WP, ABC1, ABC2, ABC, ABC_WP, OPTIMIZED, OPTIMIZED_WP, CUSTOM };

	private BufferedWriter sd1File;
	private BufferedWriter sd2File;
	private ArrayList<Sample> samples;
	private String phrasesFile;
	private boolean newPhrase = true;
	private Random r = new Random();
	private JTextField textPresented;
	private JTextField textTranscribed;
	private JLabel hitSpacebarMessage;
	private ScanningKeyboard sk;
	private String[] phrases;
	private String presentedPhrase; // presented phrase
	private StringBuilder transcribedPhrase; // transcribed phrase
	private long t1, t2, t2Last;
	private int scanSteps, scanStepsLast; // count scan steps per phrase
	private int scanStepsMin; // computed based on the presented text phrase and the type of
								// keyboard
	private double efficiency;
	int pressesOfBksp, pressesOfWordDelete;
	int scanWrapAroundsRowPhrase, scanWrapAroundsColPhrase;
	private int keyboardInt;
	private String keyboardString;
	private String keyboardDefFile;

	// setup parameters
	private String participantCode, conditionCode, sessionCode, blockCode, keyboard;
	private int phraseCount, scanInterval, scanDelay;
	private boolean showPresented, audioOn;
	private double keyboardScale;

	private Font BIG = new Font("monospaced", Font.BOLD, 24);
	private Font MEDIUM = new Font("sansserif", Font.BOLD, 16);
	private Color BACKGROUND = new Color(254, 254, 218);
	private Color FOREGROUND = new Color(11, 11, 109);

	JDialog resultsDialog;
	JOptionPane resultsPane;
	JTextArea resultsArea;

	Configuration c;

	WordPredict wordPredict;
	int wordStemIndex;
	String[] candidateWords;
	boolean wordPrediction;

	// constructor
	ScanningKeyboardExperimentGui(Configuration cArg)
	{
		// ---------------------------
		// initialize setup parameters
		// ---------------------------

		c = cArg;
		participantCode = c.getConfigurationParameter(0);
		conditionCode = c.getConfigurationParameter(1);
		sessionCode = c.getConfigurationParameter(2);
		// Note: Block code generated automatically by the software (see below)
		phraseCount = Integer.parseInt(c.getConfigurationParameter(3));
		scanInterval = Integer.parseInt(c.getConfigurationParameter(4));
		scanDelay = Integer.parseInt(c.getConfigurationParameter(5));
		phrasesFile = c.getConfigurationParameter(6);
		showPresented = Boolean.valueOf(c.getConfigurationParameter(7));
		audioOn = Boolean.valueOf(c.getConfigurationParameter(8));
		keyboard = c.getConfigurationParameter(9);
		keyboardDefFile = c.getConfigurationParameter(10);
		keyboardScale = Double.parseDouble(c.getConfigurationParameter(11));

		// remap the keyboard parameter to useful integer and string variables
		int idx = -1;
		for (int i = 0; i < ScanningKeyboardExperiment.KEYBOARD.length; ++i)
		{
			if (keyboard.equals(ScanningKeyboardExperiment.KEYBOARD[i]))
			{
				idx = i;
				break;
			}
		}

		if (idx == -1)
		{
			System.out.println("Error: can't find keyboard: " + keyboard);
			System.exit(0);
		} else
		{
			keyboardInt = MODE_ARRAY[idx];
			keyboardString = ScanningKeyboardExperiment.KEYBOARD_CODE[idx];
		}

		// -------------------------------------------
		// initialize Soft Keyboard (scanning version)
		// -------------------------------------------

		if (keyboardInt == QWERTY)
			sk = new ScanningKeyboard(qwerty, keyboardScale, scanInterval, scanDelay, audioOn,
					keyboardInt);
		else if (keyboardInt == QWERTY_WP)
			sk = new ScanningKeyboard(qwertyWP, keyboardScale, scanInterval, scanDelay, audioOn,
					keyboardInt);
		else if (keyboardInt == ABC1)
			sk = new ScanningKeyboard(abc1, keyboardScale, scanInterval, scanDelay, audioOn,
					keyboardInt);
		else if (keyboardInt == ABC2)
			sk = new ScanningKeyboard(abc2, keyboardScale, scanInterval, scanDelay, audioOn,
					keyboardInt);
		else if (keyboardInt == ABC)
			sk = new ScanningKeyboard(abc, keyboardScale, scanInterval, scanDelay, audioOn,
					keyboardInt);
		else if (keyboardInt == ABC_WP)
			sk = new ScanningKeyboard(abcWP, keyboardScale, scanInterval, scanDelay, audioOn,
					keyboardInt);
		else if (keyboardInt == OPTIMIZED)
			sk = new ScanningKeyboard(koester, keyboardScale, scanInterval, scanDelay, audioOn,
					keyboardInt);
		else if (keyboardInt == OPTIMIZED_WP)
			sk = new ScanningKeyboard(koesterWP, keyboardScale, scanInterval, scanDelay, audioOn,
					keyboardInt);
		else if (keyboardInt == CUSTOM)
			sk = new ScanningKeyboard(loadKeys(keyboardDefFile), keyboardScale, scanInterval,
					scanDelay, audioOn, keyboardInt);

		sk.addScanningKeyListener(this);

		// determine if we are using a keyboard with word prediction
		wordPrediction = false;
		for (int i = 0; i < sk.scannableKey.length; ++i)
		{
			// if any key has the label "WORD", the keyboard uses word prediction
			if (sk.scannableKey[i].getLabel().equals("WORD"))
			{
				wordPrediction = true;
				break;
			}
		}

		if (wordPrediction)
		{
			// instantiate the WordPredict object (used to generate the candidate list from a word
			// stem)
			wordPredict = new WordPredict(DICTIONARY_FILE);
			updateCandidateWords("");
			wordStemIndex = 0;
		}

		// ------------------
		// initialize phrases
		// ------------------

		Scanner inFile = null;
		try
		{
			inFile = new Scanner(new File(phrasesFile));
		} catch (FileNotFoundException e)
		{
			showError("\nPhrases file not found - " + phrasesFile);
			System.exit(1);
		}

		// read a phrase from the file and add it to an array list (repeat to end of file)
		ArrayList<String> phrasesArrayList = new ArrayList<String>();
		while (inFile.hasNextLine())
			phrasesArrayList.add(inFile.nextLine().toLowerCase());

		// convert the array list to a simple String array
		phrases = new String[phrasesArrayList.size()];
		phrases = phrasesArrayList.toArray(phrases);

		// -----------------------
		// initialize output files
		// -----------------------

		blockCode = "";
		int blockCodeNumber = 0;
		String base = ScanningKeyboardExperiment.APP_NAME + "-" + participantCode + "-"
				+ conditionCode + "-" + keyboardString + "-" + sessionCode;

		// find next available block code
		StringBuilder s1;
		StringBuilder s2;
		do
		{
			++blockCodeNumber;
			blockCode = blockCodeNumber < 10 ? "B0" + blockCodeNumber : "B" + blockCodeNumber;
			s1 = new StringBuilder();
			s2 = new StringBuilder();
			s1.append(base).append('-').append(blockCode).append(".sd1");
			s2.append(base).append('-').append(blockCode).append(".sd2");
		} while (new File(s1.toString()).exists());

		try
		{
			sd1File = new BufferedWriter(new FileWriter(s1.toString()));
			sd2File = new BufferedWriter(new FileWriter(s2.toString()));
		} catch (IOException e)
		{
			System.out.println("I/O error: can't open sd1/sd2 data files");
			System.exit(0);
		}

		try
		{
			sd1File.write(SD1_HEADER, 0, SD1_HEADER.length());
			sd1File.flush();
			sd2File.write(SD2_HEADER, 0, SD2_HEADER.length());
			sd2File.flush();
		} catch (IOException e)
		{
			System.err.println("ERROR WRITING TO DATA FILE!\n" + e);
			System.exit(1);
		}

		// -------------------------
		// initialize GUI components
		// -------------------------

		textPresented = new JTextField(50);
		textPresented.setFocusable(false);
		textPresented.setFont(BIG);
		textPresented.setBackground(new Color(254, 254, 218));
		textPresented.setForeground(new Color(11, 11, 109));

		JPanel text1Panel = new JPanel();
		text1Panel.setLayout(new FlowLayout(FlowLayout.LEFT));
		text1Panel.add(textPresented);
		text1Panel.setBorder(new TitledBorder(new EtchedBorder(), "Presented text"));

		textTranscribed = new JTextField(50);
		textTranscribed.setEditable(true); // allows I-beam to show
		textTranscribed.setFont(BIG);
		textTranscribed.setBackground(BACKGROUND);
		textTranscribed.setForeground(FOREGROUND);
		textTranscribed.addKeyListener(this);

		JPanel text2Panel = new JPanel();
		text2Panel.setLayout(new FlowLayout(FlowLayout.LEFT));
		text2Panel.add(textTranscribed);
		text2Panel.setBorder(new TitledBorder(new EtchedBorder(), "Transcribed text"));

		hitSpacebarMessage = new JLabel();
		hitSpacebarMessage.setFont(MEDIUM);
		hitSpacebarMessage.setForeground(Color.RED);
		hitSpacebarMessage.setHorizontalAlignment(SwingConstants.CENTER);
		hitSpacebarMessage.setBorder(BorderFactory.createEmptyBorder(0, 0, 20, 0));
		hitSpacebarMessage.setText(HIT_SPACEBAR_MESSAGE_TEXT);

		JPanel keyboardPanel = new JPanel();
		keyboardPanel.add(sk);
		keyboardPanel.setBorder(BorderFactory.createEmptyBorder(15, 15, 15, 15));

		// present first phrase for input
		prepareForNextPhrase();

		// prepare results dialog
		resultsArea = new JTextArea(15, 40);
		resultsArea.setFont(new Font("sansserif", Font.PLAIN, 18));
		resultsArea.setBackground((new JButton()).getBackground());
		resultsPane = new JOptionPane(resultsArea, JOptionPane.INFORMATION_MESSAGE);
		resultsPane.setFont(new Font("sansserif", Font.PLAIN, 18));
		resultsDialog = resultsPane.createDialog(this, "Information");

		// layout components
		JPanel p1 = new JPanel();
		p1.setLayout(new BoxLayout(p1, BoxLayout.Y_AXIS));
		p1.add(Box.createRigidArea(new Dimension(5, 10)));
		p1.add(text1Panel);
		p1.add(Box.createRigidArea(new Dimension(5, 10)));
		p1.add(text2Panel);
		p1.add(Box.createRigidArea(new Dimension(5, 10)));

		JPanel p = new JPanel(new BorderLayout());
		p.add(p1, "North");
		p.add(keyboardPanel, "Center");
		p.add(hitSpacebarMessage, "South");
		this.setContentPane(p);

	} // end UI constructor

	// ------------------------------------
	// implement ScanningKeyListener method
	// ------------------------------------

	@Override
	public void scanningKeyPressed(ScanningKeyboardEvent ske)
	{
		if (newPhrase)
		{
			newPhrase = false;
			if (!showPresented)
				textPresented.setText("");
		}

		String s = ske.getKeyString();

		if (s.length() == 0)
			return; // avoids a crash if the user selects a WORD key when there is no word displayed

		char c = (s.toLowerCase()).charAt(0);

		scanStepsLast = scanSteps;
		scanSteps += ske.getScanSteps();

		scanWrapAroundsRowPhrase += ske.getWrapAroundsRow();
		scanWrapAroundsColPhrase += ske.getWrapAroundsCol();

		t2Last = t2;
		t2 = System.currentTimeMillis() - t1;

		samples.add(new Sample(t2, s.toLowerCase(), ske.getScanSteps()));

		// determine the type of key pressed and process it accordingly
		switch (ske.type)
		{
			case ScanningKeyboardEvent.COMMAND: // ENTER, BKSP, or WORD_DELETE

				if (s.equals(ScanningKeyboard.ENTER)) // end of phrase
				{
					sk.stopScanning();
					sk.adjustBackgroundColors();
					newPhrase = true;

					String s1 = presentedPhrase.toLowerCase();
					String s2 = transcribedPhrase.toString().toLowerCase().trim();

					// Exclude ENTER: use last values for t2 and scanStepsTotal
					t2 = t2Last;
					scanSteps = scanStepsLast;

					scanStepsMin = minimumScanStepsPhrase(s1);
					efficiency = (double) scanStepsMin / scanSteps * 100.0;

					MSD s1s2 = new MSD(s1, s2);
					int msd = s1s2.getMSD();

					// build formated results to present in UI
					StringBuilder results = new StringBuilder();
					results.append(" DATA COLLECTED:\n");
					results.append(String.format("   Presented:\t\t%s\n", s1));
					results.append(String.format("   Transcribed:\t%s\n", s2));
					results.append(String.format("   Min scan steps:\t%d\n", scanStepsMin));
					results.append(String.format("   Scan Steps:\t%d\n", scanSteps));
					results.append(String.format("   Characters:\t\t%d (%.1f words)\n", s2.length(),
							s2.length() / 5.0));
					results.append(String.format("   Time:\t\t%.2f s (%.2f minutes)\n", t2 / 1000.0,
							t2 / 1000.0 / 60.0));
					results.append(String.format("   MSD:\t\t%d\n\n", msd));
					results.append(" PARTICIPANT PERFORMANCE:\n");
					results.append(String.format("   Entry Speed:\t%.2f wpm\n", wpm(s2, t2)));
					results.append(
							String.format("   Error rate:\t\t%.2f%%\n", s1s2.getErrorRateNew()));
					results.append(
							String.format("   SPC:\t\t%.2f\n", (double) scanSteps / s2.length()));
					results.append(String.format(
							"   Efficiency:\t\t%.1f%% (only valid if error rate is 0%%)\n",
							efficiency));
					resultsArea.setText(results.toString());
					resultsDialog.setVisible(true);

					// build output data for sd2 file
					StringBuilder sd2Stuff = new StringBuilder();
					sd2Stuff.append(ScanningKeyboardExperiment.APP_NAME).append(',');
					sd2Stuff.append(participantCode).append(',');
					sd2Stuff.append(conditionCode).append(',');
					sd2Stuff.append(keyboardString).append(',');
					sd2Stuff.append(sessionCode).append(',');
					sd2Stuff.append(blockCode).append(',');
					sd2Stuff.append(scanInterval).append(',');
					sd2Stuff.append(scanDelay).append(',');
					sd2Stuff.append(scanStepsMin).append(',');
					sd2Stuff.append(scanSteps).append(',');
					sd2Stuff.append(s2.length()).append(',');
					sd2Stuff.append(pressesOfBksp).append(',');
					sd2Stuff.append(pressesOfWordDelete).append(',');
					sd2Stuff.append(scanWrapAroundsRowPhrase).append(',');
					sd2Stuff.append(scanWrapAroundsColPhrase).append(',');
					sd2Stuff.append(t2 / 1000.0).append(','); // time in seconds
					sd2Stuff.append(msd).append(',');
					sd2Stuff.append(wpm(s2, t2)).append(',');
					sd2Stuff.append(s1s2.getErrorRateNew()).append(',');
					sd2Stuff.append((double) scanSteps / s2.length()).append(','); // SPC
					sd2Stuff.append(efficiency).append('\n');

					// build output data for sd1 file
					StringBuilder sd1Stuff = new StringBuilder();
					sd1Stuff.append(presentedPhrase).append('\n');
					sd1Stuff.append(s2).append('\n');
					for (int i = 0; i < samples.size(); ++i)
						sd1Stuff.append(samples.get(i)).append('\n');
					sd1Stuff.append('#').append('\n');

					// dump data to files
					try
					{
						sd1File.write(sd1Stuff.toString(), 0, sd1Stuff.length());
						sd1File.flush();
						sd2File.write(sd2Stuff.toString(), 0, sd2Stuff.length());
						sd2File.flush();
					} catch (IOException e)
					{
						System.err.println("ERROR WRITING TO DATA FILE!\n" + e);
						System.exit(1);
					}

					// terminate if finished last phrase in block
					if (--phraseCount == 0)
					{
						JLabel thankYou = new JLabel("End of block. Thank you.");
						thankYou.setFont(new Font("sansserif", Font.PLAIN, 16));
						JOptionPane.showMessageDialog(this, thankYou);
						try
						{
							sd1File.close();
							sd2File.close();
						} catch (IOException e)
						{
							System.err.println("ERROR CLOSING DATA FILES!\n" + e);
							System.exit(1);
						}
						System.exit(0);
					}
					prepareForNextPhrase();

				} else if (s.equals(ScanningKeyboard.BKSP)) // delete the last character
				{
					++pressesOfBksp;
					if (transcribedPhrase.length() >= 1)
					{
						transcribedPhrase.deleteCharAt(transcribedPhrase.length() - 1);
						textTranscribed.setText(transcribedPhrase.toString());
						textTranscribed.requestFocus(); // so I-beam (caret) does not disappear (is
														// this necessary?)
					}
					if (wordPrediction)
					{
						// find beginning of word stem
						wordStemIndex = transcribedPhrase.length();
						while (transcribedPhrase.length() > 0 && wordStemIndex > 0
								&& transcribedPhrase.charAt(wordStemIndex - 1) != ' ')
							--wordStemIndex;

						updateCandidateWords(transcribedPhrase.toString());
					}

				} else if (s.equals(ScanningKeyboard.WORD_DELETE))
				{
					++pressesOfWordDelete;

					String[] wordArray = transcribedPhrase.toString().split("\\s");

					// rebuild phrase, but omit the last word or word stem
					transcribedPhrase = new StringBuilder();
					for (int i = 0; i < wordArray.length - 1; ++i)
						transcribedPhrase.append(wordArray[i]).append(' ');

					textTranscribed.setText(transcribedPhrase.toString());
					textTranscribed.requestFocus(); // so I-beam (caret) does not disappear (is this
													// necessary?)
					if (wordPrediction)
					{
						wordStemIndex = transcribedPhrase.length();
						updateCandidateWords(transcribedPhrase.toString());
					}
				}
				break;

			case ScanningKeyboardEvent.CHARACTER: // just a character; add it to the transcribed
													// text

				transcribedPhrase.append(c);
				textTranscribed.setText(transcribedPhrase.toString());
				if (wordPrediction)
				{
					updateCandidateWords(transcribedPhrase.toString());
					if (c == ' ')
						wordStemIndex = transcribedPhrase.length();
				}
				break;

			case ScanningKeyboardEvent.WORD: // a word has been selected from the candidate list

				// (NOTE: wordPrediction = true)
				transcribedPhrase.delete(wordStemIndex, transcribedPhrase.length()); // delete the
																						// word stem
																						// and...
				transcribedPhrase.append(ske.getKeyString()).append(' '); // substitute the selected
																			// word (+space)
				textTranscribed.setText(transcribedPhrase.toString());
				wordStemIndex = transcribedPhrase.length();
				updateCandidateWords(transcribedPhrase.toString());
		}
	} // end scanningKeyPressed

	private void prepareForNextPhrase()
	{
		samples = new ArrayList<Sample>();
		presentedPhrase = phrases[r.nextInt(phrases.length)]; // pick random phrase from set
		transcribedPhrase = new StringBuilder();
		textPresented.setText(presentedPhrase);
		textTranscribed.setText(transcribedPhrase.toString());
		wordStemIndex = 0;
		scanSteps = 0;
		pressesOfBksp = 0;
		pressesOfWordDelete = 0;
		scanWrapAroundsRowPhrase = 0;
		scanWrapAroundsColPhrase = 0;
		hitSpacebarMessage.setText(HIT_SPACEBAR_MESSAGE_TEXT);
	}

	// compute the minimum number of scan steps required to enter the passed string
	private int minimumScanStepsString(String textString)
	{
		int min = 0;
		for (int i = 0; i < textString.length(); ++i)
		{
			String ch = textString.substring(i, i + 1);
			for (int k = 0; k < sk.scannableKey.length; ++k)
			{
				String key = sk.scannableKey[k].getLabel().toLowerCase();
				if (ch.equals(key))
				{
					min += sk.scannableKey[k].rowIdx + 1 + sk.scannableKey[k].colIdx + 1;
					break;
				}
			}
		}
		return min;
	}

	// compute the minimum scan steps required to enter the passed phrase
	private int minimumScanStepsPhrase(String phrase)
	{
		// if we're not using a keyboard with word prediction, divert to the String version of this
		// method
		if (!wordPrediction)
			return minimumScanStepsString(phrase);

		String[] wordArray = phrase.split("\\s");
		int min = 0;

		for (int i = 0; i < wordArray.length; ++i)
		{
			String word = wordArray[i];
			StringBuilder stem = new StringBuilder();
			boolean match = false;
			/*
			 * The setup for the following loop looks a bit odd, but there's a method to the
			 * madness. The loop begins with j = -1 because the first pass through the loop checks
			 * for candidates against an empty stem. Only on the second pass through the loop are
			 * characters extracted from the word and added to the word stem.
			 */
			for (int j = -1; j < word.length(); ++j)
			{
				// Note: word stem is empty on the first pass through the loop
				if (j >= 0)
					stem.append(word.charAt(j));

				// look for candidates matching the current word stem
				String[] candidateWords = wordPredict.getWords(stem.toString(),
						sk.wordIndex.length);

				// if there are no candidates, try again with the next character added to the word
				// stem
				if (candidateWords == null)
					continue;

				// we've got some candidates, so check if any of them match the word
				for (int k = 0; k < candidateWords.length; ++k)
				{
					String nextCandidate = candidateWords[k];
					if (word.equals(nextCandidate))
					{
						min += minimumScanStepsString(stem.toString());
						min += sk.scannableKey[sk.wordIndex[k]].rowIdx + 1
								+ sk.scannableKey[sk.wordIndex[k]].colIdx + 1;
						match = true;
						break;
					}
				}
				if (match)
					break;
			}
			if (!match)
			{
				min += minimumScanStepsString(word);
				if (i < wordArray.length - 1)
					min += minimumScanStepsString(" "); // include scan steps for SPACE for all
														// words except last
			}
		}
		return min;
	}

	void updateCandidateWords(String transcribedText)
	{
		String[] words = transcribedText.split("\\s");

		if (transcribedText.length() == 0
				|| transcribedText.charAt(transcribedText.length() - 1) == ' ' || words.length == 0)
			candidateWords = wordPredict.getWords("", sk.wordIndex.length);
		else
			candidateWords = wordPredict.getWords(words[words.length - 1], sk.wordIndex.length);

		// make sure the candidate word list is of the proper length (fill with "", if necessary)
		if (candidateWords == null || candidateWords.length < sk.wordIndex.length)
		{
			String[] temp = new String[sk.wordIndex.length];
			for (int i = 0; i < temp.length; ++i)
				if (candidateWords != null && i < candidateWords.length)
					temp[i] = candidateWords[i];
				else
					temp[i] = "";
			candidateWords = temp;
		}

		sk.updateCandidates(candidateWords);
	}

	// ---------------------------------
	// implement KeyListener methods (3)
	// ---------------------------------

	@Override
	public void keyPressed(KeyEvent ke)
	{
		if (KeyStroke.getKeyStroke(ke.getKeyCode(), ke.getModifiers()) == SPACEBAR)
		{
			if (!sk.scanningStarted)
			{
				sk.startScanning(); // once only at beginning of a phrase
				t1 = System.currentTimeMillis(); // start timing the input of the phrase
				hitSpacebarMessage.setText("");
			} else
				sk.switchActivated();
		}
	}

	@Override
	public void keyReleased(KeyEvent ke)
	{
		// klugy way to prevent the I-beam from advancing when the SPACE bar is pressed during
		// row-scanning
		textTranscribed.setText(
				textTranscribed.getText().substring(0, textTranscribed.getText().length() - 1));
	}

	@Override
	public void keyTyped(KeyEvent ke)
	{
	}

	// load a Key array from a custom keyboard definition file
	ScannableKey[] loadKeys(String fileArg)
	{
		Scanner customKeyFile = null;

		try
		{
			customKeyFile = new Scanner(new File(fileArg));
		} catch (FileNotFoundException e)
		{
			JLabel tmp = new JLabel("File not found: " + fileArg);
			tmp.setFont(F14);
			JOptionPane.showMessageDialog(this, tmp, "Error", JOptionPane.ERROR_MESSAGE);
			return null;
		}
		String line;
		ArrayList<ScannableKey> arrayList = new ArrayList<ScannableKey>();
		int lineNumber = 0;
		while (customKeyFile.hasNextLine())
		{
			// get next data line
			line = customKeyFile.nextLine();

			// skip blank lines or lines beginning with "#"
			if (line.length() == 0 || line.charAt(0) == '#')
				continue;

			++lineNumber;
			String[] s = line.split(",");
			if (s.length != 7)
			{
				System.out.printf("Error reading custom keyboard at line %d : %s\n", lineNumber,
						line);
				System.exit(0);
			}

			String keyText = s[0].trim();
			int rowIdx = Integer.parseInt(s[1].trim());
			int colIdx = Integer.parseInt(s[2].trim());
			double xFactor = Double.parseDouble(s[3].trim());
			double yFactor = Double.parseDouble(s[4].trim());
			double widthFactor = Double.parseDouble(s[5].trim());
			double heightFactor = Double.parseDouble(s[6].trim());
			arrayList.add(new ScannableKey(keyText, rowIdx, colIdx, xFactor, yFactor, widthFactor,
					heightFactor));
		}
		customKeyFile.close();
		ScannableKey[] keyArray = new ScannableKey[arrayList.size()];
		keyArray = arrayList.toArray(keyArray);
		return keyArray;
	}

	// popup a dialog with an error message
	void showError(String msg)
	{
		JOptionPane.showMessageDialog(null, msg, "I/O Error", JOptionPane.ERROR_MESSAGE);
	}

	// compute typing speed in wpm given text entered and time in ms
	public static double wpm(String text, long msTime)
	{
		return text.length() / (msTime / 1000.0) * (60 / 5);
	}

	// simple class to hold the information about a single character entered on the scanning
	// keyboard
	private class Sample
	{
		private long time;
		private String keyLabel;
		private int scanStepsCharacter;

		Sample(long t, String k, int scanStepsArg)
		{
			time = t;
			keyLabel = k;
			scanStepsCharacter = scanStepsArg;
		}

		public String toString()
		{
			if (keyLabel.equals(" ")) // remap SPACE character to underscore
				keyLabel = "_";

			if (keyLabel.length() > 1) // remap commands to uppercase
				keyLabel = keyLabel.toUpperCase();

			return keyLabel + ", " + time + ", " + scanStepsCharacter;
		}
	}

	// =================
	// SCANNING KEYBOARD
	// =================

	class ScanningKeyboard extends JPanel implements ActionListener
	{
		// the following avoids a "warning" with Java 1.5.0 complier (?)
		static final long serialVersionUID = 42L;

		final int DEFAULT_KEY_WIDTH = 50;
		final int DEFAULT_KEY_HEIGHT = 45;
		final int DEFAULT_FONT_SIZE = 16;

		final Color DARK_YELLOW = new Color(0x888800);
		final Color DARK_BLUE = new Color(11, 11, 109);

		final static String SPACE = "Space";
		final static String ENTER = "Enter";
		final static String BKSP = "Bksp";
		final static String WORD_DELETE = "Word_Delete";

		final int ROW_SCANNING = 100;
		final int COL_SCANNING = 200;

		JButton[] keyButton;
		int mode;
		boolean scanningStarted;
		boolean audioOn;

		ScannableKey[] scannableKey;
		Timer scanTimer;
		int scanInterval, scanDelay;
		int activeRow, activeCol;
		int activeKey;
		int scanMode; // 0 = row scanning, 1 = col scanning
		int numberOfRows;
		int scanStepsPerCharacter;
		int wrapAroundsRow, wrapAroundsCol; // count scanning wraps around at end of rows/columns

		double keyboardScale;
		int keyWidth, keyHeight;

		boolean usesWordPrediction;
		int[] wordIndex;

		ScanningKeyboardListener scanningKeyListener;

		Clip tickSound1, tickSound2, spacebarSound;

		public ScanningKeyboard(ScannableKey[] keyArg, double keyboardScaleArg, int scanIntervalArg,
				int scanDelayArg, boolean audioOnArg, int modeArg)
		{
			keyboardScale = keyboardScaleArg;
			keyWidth = (int) (keyboardScale * DEFAULT_KEY_WIDTH);
			keyHeight = (int) (keyboardScale * DEFAULT_KEY_HEIGHT);

			scanTimer = new Timer(scanIntervalArg, this);
			scanTimer.setInitialDelay(scanIntervalArg + scanDelayArg);
			scanTimer.stop();
			scanningStarted = false;

			scanInterval = scanIntervalArg;
			scanDelay = scanDelayArg;

			audioOn = audioOnArg;

			activeRow = 0;
			activeCol = 0;
			activeKey = 0;
			scanMode = ROW_SCANNING;
			scanStepsPerCharacter = 0;
			wrapAroundsRow = 0;
			wrapAroundsCol = 0;

			scannableKey = keyArg;
			keyButton = new JButton[keyArg.length];
			mode = modeArg;

			// determine number of rows
			numberOfRows = 0;
			for (int i = 0; i < scannableKey.length; ++i)
				if (scannableKey[i].rowIdx > numberOfRows)
					numberOfRows = scannableKey[i].rowIdx;
			++numberOfRows;
			// System.out.println("Number of rows = " + numberOfRows);

			// System.out.println("scannableKey.length = " + scannableKey.length);

			// determine required dimension for panel
			int width = 0;
			int height = 0;
			for (int i = 0; i < scannableKey.length; ++i)
			{
				int edgeRight = (int) (scannableKey[i].xFactor * keyWidth)
						+ (int) (scannableKey[i].widthFactor * keyWidth);
				width = edgeRight > width ? edgeRight : width;

				int edgeBottom = scannableKey[i].rowIdx
						* (int) (scannableKey[i].heightFactor * keyHeight)
						+ (int) (scannableKey[i].heightFactor * keyHeight);
				height = edgeBottom > height ? edgeBottom : height;
			}
			Dimension d = new Dimension(width, height);

			// System.out.println("width=" + width + ", height=" + height);
			// System.out.println("keyButton.length=" + keyButton.length);

			// initialize the keys (instances of JButton)
			for (int i = 0; i < keyButton.length; ++i)
			{
				keyButton[i] = new JButton(scannableKey[i].getLabel());
				keyButton[i].setMargin(new Insets(0, 0, 0, 0));

				// set font sizes
				if (scannableKey[i].getLabel().equals("WORD"))
					keyButton[i].setFont(new Font("sannsserif", Font.BOLD,
							(int) (0.8 * keyboardScale * DEFAULT_FONT_SIZE)));
				else
					keyButton[i].setFont(new Font("sannsserif", Font.BOLD,
							(int) (keyboardScale * DEFAULT_FONT_SIZE)));

				keyButton[i].setBounds((int) (scannableKey[i].xFactor * keyWidth),
						(int) (scannableKey[i].yFactor * keyHeight),
						(int) (scannableKey[i].widthFactor * keyWidth),
						(int) (scannableKey[i].heightFactor * keyHeight));
				this.add(keyButton[i]);
			}

			// determine if the keyboard uses word prediction and, if so, fill an array with the key
			// indices
			usesWordPrediction = false;
			ArrayList<Integer> indexList = new ArrayList<Integer>();
			for (int i = 0; i < scannableKey.length; ++i)
			{
				if (scannableKey[i].getLabel().equals("WORD"))
				{
					usesWordPrediction = true;
					indexList.add(new Integer(i));
				}
			}
			if (usesWordPrediction)
			{

				wordIndex = new int[indexList.size()];
				for (int i = 0; i < wordIndex.length; ++i)
				{
					wordIndex[i] = Integer.valueOf(indexList.get(i));

					// while we're at it, change the color of the word on the key
					keyButton[Integer.valueOf(indexList.get(i))].setForeground(DARK_BLUE);
				}
			}

			// initialize sounds (if audio on)
			if (audioOn)
			{
				tickSound1 = initSound("tickRowScanning.wav");
				tickSound2 = initSound("tickColScanning.wav");
				spacebarSound = initSound("switch.wav");
			}

			adjustBackgroundColors();
			this.setLayout(null);
			this.setPreferredSize(d);
			this.setMaximumSize(d);
		} // end constructor

		/*
		 * Update the list of candidate words (word prediction only) based on either an empty word
		 * or a word stem. Note that the wordIndex array holds the locations (indices) of the keys
		 * that present the candidate words.
		 */
		public void updateCandidates(String[] s)
		{
			for (int i = 0; i < s.length; ++i)
				keyButton[wordIndex[i]].setText(s[i]);
		}

		public void startScanning()
		{
			scanningStarted = true;
			scanTimer.start();
			t1 = 0;
			adjustBackgroundColors();
		}

		public void stopScanning()
		{
			scanningStarted = false;
			scanTimer.stop();
		}

		// method used in app to attach listener for scanning key events
		public void addScanningKeyListener(ScanningKeyboardListener scanningKeyListenerArg)
		{
			scanningKeyListener = scanningKeyListenerArg;
		}

		// do this with each timeout of the scan timer
		public void actionPerformed(ActionEvent ae)
		{
			++scanStepsPerCharacter;

			// advance to next row or column
			if (scanMode == ROW_SCANNING)
			{
				if (audioOn)
					playSound(tickSound1);
				advanceToNextRow();
			} else if (scanMode == COL_SCANNING)
			{
				if (audioOn)
					playSound(tickSound2);
				advanceToNextCol();
			}
			adjustBackgroundColors();
		}

		private void advanceToNextRow()
		{
			++activeRow; // advance to next row
			if (activeRow == numberOfRows) // if past bottom row, ...
			{
				++wrapAroundsRow; // increment the count for row wrap-arounds (indicates a missed
									// opportunity)
				activeRow = 0; // return to top row
			}

			activeCol = 0;

			// find index of first key in active row (skip over filler keys, if any)
			for (int i = 0; i < scannableKey.length; ++i)
			{
				if (scannableKey[i].rowIdx == activeRow && scannableKey[i].colIdx == 0)
				{
					activeKey = i;
					break;
				}
			}
		}

		private void advanceToNextCol()
		{
			// first, just try advancing the pointers forward
			++activeCol;
			++activeKey;
			
			//System.out.println("activeCol=" + activeCol + ", activeKey=" + activeKey + ", scannableKey.length=" + scannableKey.length);
			
			// skip keys with no text
			while (activeKey != scannableKey.length && scannableKey[activeKey].getLabel().equals(""))
			{
				//System.out.println("Got here!");
				++activeCol;
				++activeKey;
			}

			/*
			 * Check if we are past the end of the keys or if we've advanced past the last key in
			 * the row. If so, set the active key to zero and...
			 */
			if (activeKey == scannableKey.length || scannableKey[activeKey].colIdx != activeCol)
			{
				++wrapAroundsCol;
				activeKey = 0;

				// find the first key at beginning of the active row
				while (activeRow != scannableKey[activeKey].rowIdx)
					++activeKey;

				activeCol = 0;
				while (scannableKey[activeKey].colIdx == -1)
					++activeKey;
			}
		}

		// adjust key backgrounds, as per the currently active row or column and the scanning mode
		private void adjustBackgroundColors()
		{
			for (int i = 0; i < keyButton.length; ++i)
			{
				if (scanMode == ROW_SCANNING)
				{
					if (scanningStarted && scannableKey[i].rowIdx == activeRow)
						if (scannableKey[i].colIdx == -1)
							keyButton[i].setBackground(DARK_YELLOW);
						else
							keyButton[i].setBackground(Color.YELLOW);
					else if (scannableKey[i].colIdx == -1)
						keyButton[i].setBackground(Color.GRAY);
					else
						keyButton[i].setBackground(Color.LIGHT_GRAY);
				} else if (scanMode == COL_SCANNING)
				{
					if (scanningStarted && i == activeKey)
						keyButton[i].setBackground(Color.YELLOW);
					else if (scannableKey[i].colIdx == -1)
						keyButton[i].setBackground(Color.GRAY);
					else
						keyButton[i].setBackground(Color.LIGHT_GRAY);
				}
			}
		}

		// the SPACEBAR has been pressed (called from main app in keyPressed method)
		public void switchActivated()
		{
			++scanStepsPerCharacter; // inc scanSteps because switch activations occur before the
										// end of a timeout

			if (audioOn)
				playSound(spacebarSound);

			switch (scanMode)
			{
				case ROW_SCANNING: // Switch to col scanning
					// find first selectable key in active row
					activeKey = 0;
					while (activeRow != scannableKey[activeKey].rowIdx
							|| scannableKey[activeKey].colIdx == -1)
						++activeKey;
					break;

				case COL_SCANNING: // a character has been entered; send it to the app, then return
									// to row scanning

					// first determine the "type" of key that was selected
					int type = -1;
					String label = scannableKey[activeKey].getLabel();
					if (label.equals("WORD"))
						type = ScanningKeyboardEvent.WORD;
					else if (label.equals(ENTER) || label.equals(BKSP) || label.equals(WORD_DELETE))
						type = ScanningKeyboardEvent.COMMAND;
					else if (label.length() > 0)
						type = ScanningKeyboardEvent.CHARACTER;

					// if a word was selected, replace the key label with the candidate word
					if (type == ScanningKeyboardEvent.WORD)
						label = sk.keyButton[activeKey].getText();

					ScanningKeyboardEvent ske = new ScanningKeyboardEvent(label,
							System.currentTimeMillis(), scanStepsPerCharacter, wrapAroundsRow,
							wrapAroundsCol, type);
					scanningKeyListener.scanningKeyPressed(ske);

					activeRow = 0;
					activeCol = 0;
					scanStepsPerCharacter = 0;
					wrapAroundsRow = 0;
					wrapAroundsCol = 0;
			}

			// advance to col scanning or return to row scanning
			scanMode = scanMode == ROW_SCANNING ? COL_SCANNING : ROW_SCANNING;

			adjustBackgroundColors();
			if (scanningStarted)
				scanTimer.restart();
		}
	} // end ScanningKeyboard class definition

	ScannableKey[] qwerty =
	{

			// 1st row (top)
			new ScannableKey("Q", 0, 0, 0, 0, 1, 1), new ScannableKey("W", 0, 1, 1, 0, 1, 1),
			new ScannableKey("E", 0, 2, 2, 0, 1, 1), new ScannableKey("R", 0, 3, 3, 0, 1, 1),
			new ScannableKey("T", 0, 4, 4, 0, 1, 1), new ScannableKey("Y", 0, 5, 5, 0, 1, 1),
			new ScannableKey("U", 0, 6, 6, 0, 1, 1), new ScannableKey("I", 0, 7, 7, 0, 1, 1),
			new ScannableKey("O", 0, 8, 8, 0, 1, 1), new ScannableKey("P", 0, 9, 9, 0, 1, 1),
			new ScannableKey(ScanningKeyboard.BKSP, 0, 10, 10, 0, (int) (7f / 3), 1),

			// 2nd row
			new ScannableKey("", 1, -1, 0, 1, (1.0 / 3), 1),
			new ScannableKey("A", 1, 0, (1.0 / 3) + 0, 1, 1, 1),
			new ScannableKey("S", 1, 1, (1.0 / 3) + 1, 1, 1, 1),
			new ScannableKey("D", 1, 2, (1.0 / 3) + 2, 1, 1, 1),
			new ScannableKey("F", 1, 3, (1.0 / 3) + 3, 1, 1, 1),
			new ScannableKey("G", 1, 4, (1.0 / 3) + 4, 1, 1, 1),
			new ScannableKey("H", 1, 5, (1.0 / 3) + 5, 1, 1, 1),
			new ScannableKey("J", 1, 6, (1.0 / 3) + 6, 1, 1, 1),
			new ScannableKey("K", 1, 7, (1.0 / 3) + 7, 1, 1, 1),
			new ScannableKey("L", 1, 8, (1.0 / 3) + 8, 1, 1, 1),
			new ScannableKey(ScanningKeyboard.ENTER, 1, 9, (1.0 / 3) + 9, 1, 2, 1),
			new ScannableKey("", 1, -1, (1.0 / 3) + 11, 1, (2.05 / 3), 1),

			// 3rd row
			new ScannableKey("", 2, -1, 0, 2, 2.0 / 3, 1),
			new ScannableKey("Z", 2, 0, (2.0 / 3) + 0, 2, 1, 1),
			new ScannableKey("X", 2, 1, (2.0 / 3) + 1, 2, 1, 1),
			new ScannableKey("C", 2, 2, (2.0 / 3) + 2, 2, 1, 1),
			new ScannableKey("V", 2, 3, (2.0 / 3) + 3, 2, 1, 1),
			new ScannableKey("B", 2, 4, (2.0 / 3) + 4, 2, 1, 1),
			new ScannableKey("N", 2, 5, (2.0 / 3) + 5, 2, 1, 1),
			new ScannableKey("M", 2, 6, (2.0 / 3) + 6, 2, 1, 1),
			new ScannableKey("", 2, -1, (2.0 / 3) + 7, 2, 13.05 / 3, 1),

			// 4th row (bottom)
			new ScannableKey("", 3, -1, 0, 3, 2, 1),
			new ScannableKey(ScanningKeyboard.SPACE, 3, 0, 2, 3, 6, 1),
			new ScannableKey("", 3, -1, 8, 3, 4, 1) };

	ScannableKey[] qwertyWP =
	{
			// 1st row (top)
			new ScannableKey("WORD", 0, 0, (0f * 37f / 18), 0, (37f / 18), 1),
			new ScannableKey("WORD", 0, 1, (1f * 37f / 18), 0, (37f / 18), 1),
			new ScannableKey("WORD", 0, 2, (2f * 37f / 18), 0, (37f / 18), 1),
			new ScannableKey("WORD", 0, 3, (3f * 37f / 18), 0, (37f / 18), 1),
			new ScannableKey("WORD", 0, 4, (4f * 37f / 18), 0, (37f / 18), 1),
			new ScannableKey("WORD", 0, 5, (5f * 37f / 18), 0, (37.2f / 18), 1),

			// 2nd row
			new ScannableKey("Q", 1, 0, 0, 1, 1, 1), new ScannableKey("W", 1, 1, 1, 1, 1, 1),
			new ScannableKey("E", 1, 2, 2, 1, 1, 1), new ScannableKey("R", 1, 3, 3, 1, 1, 1),
			new ScannableKey("T", 1, 4, 4, 1, 1, 1), new ScannableKey("Y", 1, 5, 5, 1, 1, 1),
			new ScannableKey("U", 1, 6, 6, 1, 1, 1), new ScannableKey("I", 1, 7, 7, 1, 1, 1),
			new ScannableKey("O", 1, 8, 8, 1, 1, 1), new ScannableKey("P", 1, 9, 9, 1, 1, 1),
			new ScannableKey(ScanningKeyboard.BKSP, 1, 10, 10, 1, (7f / 3), 1),

			// 3rd row
			new ScannableKey("", 2, -1, 0, 2, (1 / 3f), 1),
			new ScannableKey("A", 2, 0, (1 / 3f) + 0, 2, 1, 1),
			new ScannableKey("S", 2, 1, (1 / 3f) + 1, 2, 1, 1),
			new ScannableKey("D", 2, 2, (1 / 3f) + 2, 2, 1, 1),
			new ScannableKey("F", 2, 3, (1 / 3f) + 3, 2, 1, 1),
			new ScannableKey("G", 2, 4, (1 / 3f) + 4, 2, 1, 1),
			new ScannableKey("H", 2, 5, (1 / 3f) + 5, 2, 1, 1),
			new ScannableKey("J", 2, 6, (1 / 3f) + 6, 2, 1, 1),
			new ScannableKey("K", 2, 7, (1 / 3f) + 7, 2, 1, 1),
			new ScannableKey("L", 2, 8, (1 / 3f) + 8, 2, 1, 1),
			new ScannableKey(ScanningKeyboard.ENTER, 2, 9, (1 / 3f) + 9, 2, 2, 1),
			new ScannableKey("", 2, -1, (1 / 3f) + 11, 2, (3f / 3), 1),

			// 4th row
			new ScannableKey("", 3, -1, 0, 3, (2.0 / 3), 1),
			new ScannableKey("Z", 3, 0, (2 / 3.0) + 0, 3, 1, 1),
			new ScannableKey("X", 3, 1, (2 / 3.0) + 1, 3, 1, 1),
			new ScannableKey("C", 3, 2, (2 / 3.0) + 2, 3, 1, 1),
			new ScannableKey("V", 3, 3, (2 / 3.0) + 3, 3, 1, 1),
			new ScannableKey("B", 3, 4, (2 / 3.0) + 4, 3, 1, 1),
			new ScannableKey("N", 3, 5, (2 / 3.0) + 5, 3, 1, 1),
			new ScannableKey("M", 3, 6, (2 / 3.0) + 6, 3, 1, 1),
			new ScannableKey(ScanningKeyboard.WORD_DELETE, 3, 7, (2 / 3.0) + 7, 3, 7.0 / 3, 1),
			new ScannableKey("", 3, -1, (2 / 3.0) + 9, 3, (8.0 / 3), 1),

			// 5th row (bottom)
			new ScannableKey("", 4, -1, 0, 4, 2, 1),
			new ScannableKey(ScanningKeyboard.SPACE, 4, 0, 2, 4, 6, 1),
			new ScannableKey("", 4, -1, 8, 4, (13f / 3), 1) };

	ScannableKey[] abc1 =
	{
			// 1st row (top)
			new ScannableKey(ScanningKeyboard.SPACE, 0, 0, 0, 0, 2, 1),
			new ScannableKey(ScanningKeyboard.BKSP, 0, 1, 2, 0, 2, 1),
			new ScannableKey(ScanningKeyboard.ENTER, 0, 2, 4, 0, 2, 1),

			// 2nd row
			new ScannableKey("A", 1, 0, 0, 1, 1, 1), new ScannableKey("B", 1, 1, 1, 1, 1, 1),
			new ScannableKey("C", 1, 2, 2, 1, 1, 1), new ScannableKey("D", 1, 3, 3, 1, 1, 1),
			new ScannableKey("E", 1, 4, 4, 1, 1, 1), new ScannableKey("F", 1, 5, 5, 1, 1, 1),

			// 3rd row
			new ScannableKey("G", 2, 0, 0, 2, 1, 1), new ScannableKey("H", 2, 1, 1, 2, 1, 1),
			new ScannableKey("I", 2, 2, 2, 2, 1, 1), new ScannableKey("J", 2, 3, 3, 2, 1, 1),
			new ScannableKey("K", 2, 4, 4, 2, 1, 1), new ScannableKey("L", 2, 5, 5, 2, 1, 1),

			// 4th row
			new ScannableKey("M", 3, 0, 0, 3, 1, 1), new ScannableKey("N", 3, 1, 1, 3, 1, 1),
			new ScannableKey("O", 3, 2, 2, 3, 1, 1), new ScannableKey("P", 3, 3, 3, 3, 1, 1),
			new ScannableKey("Q", 3, 4, 4, 3, 1, 1), new ScannableKey("R", 3, 5, 5, 3, 1, 1),

			// 5th row
			new ScannableKey("S", 4, 0, 0, 4, 1, 1), new ScannableKey("T", 4, 1, 1, 4, 1, 1),
			new ScannableKey("U", 4, 2, 2, 4, 1, 1), new ScannableKey("V", 4, 3, 3, 4, 1, 1),
			new ScannableKey("W", 4, 4, 4, 4, 1, 1), new ScannableKey("X", 4, 5, 5, 4, 1, 1),

			// 6th row (bottom)
			new ScannableKey("Y", 5, 0, 0, 5, 1, 1), new ScannableKey("Z", 5, 1, 1, 5, 1, 1),
			new ScannableKey("", 5, -1, 2, 5, 1, 1), new ScannableKey("", 5, -1, 3, 5, 1, 1),
			new ScannableKey("", 5, -1, 4, 5, 1, 1), new ScannableKey("", 5, -1, 5, 5, 1, 1), };

	ScannableKey[] abc2 =
	{
			// 1st row (top)
			new ScannableKey("A", 0, 0, 0, 0, 1, 1), new ScannableKey("B", 0, 1, 1, 0, 1, 1),
			new ScannableKey("C", 0, 2, 2, 0, 1, 1), new ScannableKey("D", 0, 3, 3, 0, 1, 1),
			new ScannableKey("E", 0, 4, 4, 0, 1, 1), new ScannableKey("F", 0, 5, 5, 0, 1, 1),

			// 2nd row
			new ScannableKey("G", 1, 0, 0, 1, 1, 1), new ScannableKey("H", 1, 1, 1, 1, 1, 1),
			new ScannableKey("I", 1, 2, 2, 1, 1, 1), new ScannableKey("J", 1, 3, 3, 1, 1, 1),
			new ScannableKey("K", 1, 4, 4, 1, 1, 1), new ScannableKey("L", 1, 5, 5, 1, 1, 1),

			// 3rd row
			new ScannableKey("M", 2, 0, 0, 2, 1, 1), new ScannableKey("N", 2, 1, 1, 2, 1, 1),
			new ScannableKey("O", 2, 2, 2, 2, 1, 1), new ScannableKey("P", 2, 3, 3, 2, 1, 1),
			new ScannableKey("Q", 2, 4, 4, 2, 1, 1), new ScannableKey("R", 2, 5, 5, 2, 1, 1),

			// 4th row
			new ScannableKey("S", 3, 0, 0, 3, 1, 1), new ScannableKey("T", 3, 1, 1, 3, 1, 1),
			new ScannableKey("U", 3, 2, 2, 3, 1, 1), new ScannableKey("V", 3, 3, 3, 3, 1, 1),
			new ScannableKey("W", 3, 4, 4, 3, 1, 1), new ScannableKey("X", 3, 5, 5, 3, 1, 1),

			// 5th row
			new ScannableKey("Y", 4, 0, 0, 4, 1, 1), new ScannableKey("Z", 4, 1, 1, 4, 1, 1),
			new ScannableKey("", 4, -1, 2, 4, 1, 1), new ScannableKey("", 4, -1, 3, 4, 1, 1),
			new ScannableKey("", 4, -1, 4, 4, 1, 1), new ScannableKey("", 4, -1, 5, 4, 1, 1),

			// 6th row (bottom)
			new ScannableKey(ScanningKeyboard.SPACE, 5, 0, 0, 5, 2 * 1, 1),
			new ScannableKey(ScanningKeyboard.BKSP, 5, 1, 2, 5, 2 * 1, 1),
			new ScannableKey(ScanningKeyboard.ENTER, 5, 2, 4, 5, 2 * 1, 1), };

	ScannableKey[] abc =
	{
			// 1st row (top)
			new ScannableKey("A", 0, 0, 0, 0, 1, 1), new ScannableKey("B", 0, 1, 1, 0, 1, 1),
			new ScannableKey("C", 0, 2, 2, 0, 1, 1), new ScannableKey("D", 0, 3, 3, 0, 1, 1),
			new ScannableKey("E", 0, 4, 4, 0, 1, 1), new ScannableKey("F", 0, 5, 5, 0, 1, 1),
			new ScannableKey("G", 0, 6, 6, 0, 1, 1), new ScannableKey("H", 0, 7, 7, 0, 1, 1),
			new ScannableKey("I", 0, 8, 8, 0, 1, 1), new ScannableKey("J", 0, 9, 9, 0, 1, 1),
			new ScannableKey("K", 0, 10, 10, 0, 1, 1), new ScannableKey("L", 0, 11, 11, 0, 1, 1),
			new ScannableKey("M", 0, 12, 12, 0, 1, 1),

			// 2nd row
			new ScannableKey("N", 1, 0, 0, 1, 1, 1), new ScannableKey("O", 1, 1, 1, 1, 1, 1),
			new ScannableKey("P", 1, 2, 2, 1, 1, 1), new ScannableKey("Q", 1, 3, 3, 1, 1, 1),
			new ScannableKey("R", 1, 4, 4, 1, 1, 1), new ScannableKey("S", 1, 5, 5, 1, 1, 1),
			new ScannableKey("T", 1, 6, 6, 1, 1, 1), new ScannableKey("U", 1, 7, 7, 1, 1, 1),
			new ScannableKey("V", 1, 8, 8, 1, 1, 1), new ScannableKey("W", 1, 9, 9, 1, 1, 1),
			new ScannableKey("X", 1, 10, 10, 1, 1, 1), new ScannableKey("Y", 1, 11, 11, 1, 1, 1),
			new ScannableKey("Z", 1, 12, 12, 1, 1, 1),

			// 3rd row (bottom)
			new ScannableKey(ScanningKeyboard.SPACE, 2, 0, 0 * (13.0 / 4), 2, (13.0 / 4), 1),
			new ScannableKey(ScanningKeyboard.BKSP, 2, 1, 1 * (13.0 / 4), 2, (13.0 / 4), 1),
			new ScannableKey(ScanningKeyboard.WORD_DELETE, 2, 2, 2 * (13.0 / 4), 2, (13.0 / 4), 1),
			new ScannableKey(ScanningKeyboard.ENTER, 2, 3, 3 * (13.0 / 4), 2, (13.1 / 4), 1) };

	ScannableKey[] abcWP =
	{
			// 1st row (top)
			new ScannableKey("A", 0, 0, 0, 0, 1, 1), new ScannableKey("B", 0, 1, 1, 0, 1, 1),
			new ScannableKey("C", 0, 2, 2, 0, 1, 1), new ScannableKey("D", 0, 3, 3, 0, 1, 1),
			new ScannableKey("E", 0, 4, 4, 0, 1, 1), new ScannableKey("F", 0, 5, 5, 0, 1, 1),
			new ScannableKey("G", 0, 6, 6, 0, 1, 1), new ScannableKey("H", 0, 7, 7, 0, 1, 1),
			new ScannableKey("I", 0, 8, 8, 0, 1, 1), new ScannableKey("J", 0, 9, 9, 0, 1, 1),
			new ScannableKey("K", 0, 10, 10, 0, 1, 1), new ScannableKey("L", 0, 11, 11, 0, 1, 1),
			new ScannableKey("M", 0, 12, 12, 0, 1, 1),

			// 2nd row
			new ScannableKey("N", 1, 0, 0, 1, 1, 1), new ScannableKey("O", 1, 1, 1, 1, 1, 1),
			new ScannableKey("P", 1, 2, 2, 1, 1, 1), new ScannableKey("Q", 1, 3, 3, 1, 1, 1),
			new ScannableKey("R", 1, 4, 4, 1, 1, 1), new ScannableKey("S", 1, 5, 5, 1, 1, 1),
			new ScannableKey("T", 1, 6, 6, 1, 1, 1), new ScannableKey("U", 1, 7, 7, 1, 1, 1),
			new ScannableKey("V", 1, 8, 8, 1, 1, 1), new ScannableKey("W", 1, 9, 9, 1, 1, 1),
			new ScannableKey("X", 1, 10, 10, 1, 1, 1), new ScannableKey("Y", 1, 11, 11, 1, 1, 1),
			new ScannableKey("Z", 1, 12, 12, 1, 1, 1),

			// 3rd row
			new ScannableKey("WORD", 2, 0, 0 * (13.0 / 6), 2, (13.0 / 6), 1),
			new ScannableKey("WORD", 2, 1, 1 * (13.0 / 6), 2, (13.0 / 6), 1),
			new ScannableKey("WORD", 2, 2, 2 * (13.0 / 6), 2, (13.0 / 6), 1),
			new ScannableKey("WORD", 2, 3, 3 * (13.0 / 6), 2, (13.0 / 6), 1),
			new ScannableKey("WORD", 2, 4, 4 * (13.0 / 6), 2, (13.0 / 6), 1),
			new ScannableKey("WORD", 2, 5, 5 * (13.0 / 6), 2, (13.1 / 6), 1),

			// 4th row (bottom)
			new ScannableKey(ScanningKeyboard.SPACE, 3, 0, 0 * (13.0 / 4), 3, (13.0 / 4), 1),
			new ScannableKey(ScanningKeyboard.BKSP, 3, 1, 1 * (13.0 / 4), 3, (13.0 / 4), 1),
			new ScannableKey(ScanningKeyboard.WORD_DELETE, 3, 2, 2 * (13.0 / 4), 3, (13.0 / 4), 1),
			new ScannableKey(ScanningKeyboard.ENTER, 3, 3, 3 * (13.0 / 4), 3, (13.1 / 4), 1) };

	ScannableKey[] koester =
	{
			// 1st row (first)
			new ScannableKey(ScanningKeyboard.SPACE, 0, 0, 0, 0, 1, 1),
			new ScannableKey("E", 0, 1, 1, 0, 1, 1), new ScannableKey("A", 0, 2, 2, 0, 1, 1),
			new ScannableKey("R", 0, 3, 3, 0, 1, 1), new ScannableKey("D", 0, 4, 4, 0, 1, 1),
			new ScannableKey("U", 0, 5, 5, 0, 1, 1),

			// 2nd row
			new ScannableKey("T", 1, 0, 0, 1, 1, 1), new ScannableKey("N", 1, 1, 1, 1, 1, 1),
			new ScannableKey("S", 1, 2, 2, 1, 1, 1), new ScannableKey("F", 1, 3, 3, 1, 1, 1),
			new ScannableKey("W", 1, 4, 4, 1, 1, 1), new ScannableKey("B", 1, 5, 5, 1, 1, 1),

			// 3rd row
			new ScannableKey("O", 2, 0, 0, 2, 1, 1), new ScannableKey("H", 2, 1, 1, 2, 1, 1),
			new ScannableKey("C", 2, 2, 2, 2, 1, 1), new ScannableKey("P", 2, 3, 3, 2, 1, 1),
			new ScannableKey("V", 2, 4, 4, 2, 1, 1), new ScannableKey("J", 2, 5, 5, 2, 1, 1),

			// 4th row
			new ScannableKey("I", 3, 0, 0, 3, 1, 1), new ScannableKey("M", 3, 1, 1, 3, 1, 1),
			new ScannableKey("Y", 3, 2, 2, 3, 1, 1), new ScannableKey("K", 3, 3, 3, 3, 1, 1),
			new ScannableKey("Q", 3, 4, 4, 3, 1, 1), new ScannableKey("", 3, -1, 5, 3, 1, 1),

			// 5th row
			new ScannableKey("L", 4, 0, 0, 4, 1, 1), new ScannableKey("G", 4, 1, 1, 4, 1, 1),
			new ScannableKey("X", 4, 2, 2, 4, 1, 1), new ScannableKey("Z", 4, 3, 3, 4, 1, 1),
			new ScannableKey("", 4, -1, 4, 4, 1, 1), new ScannableKey("", 4, -1, 5, 4, 1, 1),

			// 6th row (last)
			new ScannableKey(ScanningKeyboard.BKSP, 5, 0, 0, 5, 3, 1),
			new ScannableKey(ScanningKeyboard.ENTER, 5, 1, 3, 5, 3, 1) };

	ScannableKey[] koesterWP =
	{
			// 1st row (first)
			new ScannableKey(ScanningKeyboard.SPACE, 0, 0, 0, 0, 1, 1),
			new ScannableKey("E", 0, 1, 1, 0, 1, 1), new ScannableKey("A", 0, 2, 2, 0, 1, 1),
			new ScannableKey("R", 0, 3, 3, 0, 1, 1), new ScannableKey("D", 0, 4, 4, 0, 1, 1),
			new ScannableKey("U", 0, 5, 5, 0, 1, 1), new ScannableKey("WORD", 0, 6, 6, 0, 2, 1),

			// 2nd row
			new ScannableKey("T", 1, 0, 0, 1, 1, 1), new ScannableKey("N", 1, 1, 1, 1, 1, 1),
			new ScannableKey("S", 1, 2, 2, 1, 1, 1), new ScannableKey("F", 1, 3, 3, 1, 1, 1),
			new ScannableKey("W", 1, 4, 4, 1, 1, 1), new ScannableKey("B", 1, 5, 5, 1, 1, 1),
			new ScannableKey("WORD", 1, 6, 6, 1, 2, 1),

			// 3rd row
			new ScannableKey("O", 2, 0, 0, 2, 1, 1), new ScannableKey("H", 2, 1, 1, 2, 1, 1),
			new ScannableKey("C", 2, 2, 2, 2, 1, 1), new ScannableKey("P", 2, 3, 3, 2, 1, 1),
			new ScannableKey("V", 2, 4, 4, 2, 1, 1), new ScannableKey("J", 2, 5, 5, 2, 1, 1),
			new ScannableKey("WORD", 2, 6, 6, 2, 2, 1),

			// 4th row
			new ScannableKey("I", 3, 0, 0, 3, 1, 1), new ScannableKey("M", 3, 1, 1, 3, 1, 1),
			new ScannableKey("Y", 3, 2, 2, 3, 1, 1), new ScannableKey("K", 3, 3, 3, 3, 1, 1),
			new ScannableKey("Q", 3, 4, 4, 3, 1, 1), new ScannableKey("", 3, -1, 5, 3, 1, 1),
			new ScannableKey("WORD", 3, 6, 6, 3, 2, 1),

			// 5th row
			new ScannableKey("L", 4, 0, 0, 4, 1, 1), new ScannableKey("G", 4, 1, 1, 4, 1, 1),
			new ScannableKey("X", 4, 2, 2, 4, 1, 1), new ScannableKey("Z", 4, 3, 3, 4, 1, 1),
			new ScannableKey("", 4, -1, 4, 4, 1, 1), new ScannableKey("", 4, -1, 5, 4, 1, 1),
			new ScannableKey("WORD", 4, 6, 6, 4, 2, 1),

			// 6th row (last)
			new ScannableKey(ScanningKeyboard.BKSP, 5, 0, 0, 5, 3, 1),
			new ScannableKey(ScanningKeyboard.ENTER, 5, 1, 3, 5, 3, 1),
			new ScannableKey("WORD", 5, 2, 6, 5, 2, 1), };

	// ====================================================
	// ScannableKey (class for keys in a scannable keyboard
	// ====================================================
	class ScannableKey
	{
		String label;
		int rowIdx;
		int colIdx; // scanning offset within row
		double xFactor; // x-offset in pixels (when multiplied by key width)
		double yFactor; // y-offset in pixels (when multiplied by key height)
		double widthFactor; // width in pixels (when multiplied by key width)
		double heightFactor; // height in pixels (when multiplied by key height)

		ScannableKey(String labelArg, int rowIdxArg, int colIdxArg, double xFactorArg,
				double yFactorArg, double widthFactorArg, double heightFactorArg)
		{
			label = labelArg;
			rowIdx = rowIdxArg;
			colIdx = colIdxArg;
			xFactor = xFactorArg;
			yFactor = yFactorArg;
			widthFactor = widthFactorArg;
			heightFactor = heightFactorArg;
		}

		String getLabel()
		{
			if (label.equals("Space"))
				return " ";
			else
				return label;
		};
	}

	// Initialize stream for sound clip
	public Clip initSound(String soundFile)
	{
		AudioInputStream audioInputStream;
		Clip c = null;
		try
		{
			// Added for executable Jar:
			audioInputStream = AudioSystem.getAudioInputStream(
					new BufferedInputStream(getClass().getResourceAsStream(soundFile)));
			c = AudioSystem.getClip();
			c.open(audioInputStream);
		} catch (Exception e)
		{
			System.out.println("ERROR: Unable to load " + soundFile);
		}
		return c;
	}

	// play sound
	public void playSound(Clip c)
	{
		if (c != null)
		{
			// sounds better if just a portion of the clip is played (based on trial and error)
			c.setFramePosition(c.getFrameLength() - 500);
			c.start(); // stops at end of clip
		}
	}
}

// ===================
// ScanningKeyListener
// ===================

interface ScanningKeyboardListener
{
	void scanningKeyPressed(ScanningKeyboardEvent ske);
}

// =====================
// ScanningKeyboardEvent
// =====================

class ScanningKeyboardEvent
{
	// some constants representing the "type" of event
	final static int CHARACTER = 100;
	final static int WORD = 200;
	final static int COMMAND = 300;

	String keyString;
	long timestamp;
	int scanSteps;
	int rowWraps, colWraps;
	int type;

	public ScanningKeyboardEvent(String keyStringArg, long timestampArg, int scanStepsArg,
			int rowWrapsArg, int colWrapsArg, int typeArg)
	{
		keyString = keyStringArg;
		timestamp = timestampArg;
		scanSteps = scanStepsArg;
		rowWraps = rowWrapsArg;
		colWraps = colWrapsArg;
		type = typeArg;
	}

	public String getKeyString()
	{
		return keyString;
	}

	public long getTimestamp()
	{
		return timestamp;
	}

	public int getScanSteps()
	{
		return scanSteps;
	}

	public int getWrapAroundsRow()
	{
		return rowWraps;
	}

	public int getWrapAroundsCol()
	{
		return colWraps;
	}

	public int getType()
	{
		return type;
	}

	public String toString()
	{
		return String.format("%s,%d,%d,%d,%d,%d", keyString, timestamp, scanSteps, rowWraps,
				colWraps, type);
	}
}
